#include "time.h"
#include "stdio.h"
#include "conio.h"
#include "string.h"

#include "MSufSort.h"
#include "BWT.h"




void LoadSourceFile(char * fileName, char * & data, int & length)
{
	// loads the source file into a buffer.
	data = 0;
	length = 0;

	FILE * fp = fopen(fileName, "rb");
	if (fp == 0)
		return;

	fseek(fp, 0, SEEK_END);
	length = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	if (length < 1)
		return;

	data = new char[length];
	fread(data, 1, length, fp);
	fclose(fp);
}





char * GetFileName(char * filename)
{
	int len = (int)strlen(filename);
	char * ret = filename + len;
	while ((ret > filename) && (ret[0] != '\\'))
		ret--;
	if (ret[0] == '\\')
		ret++;
	return ret;
}









bool IsBwtFile(char * fileName)
{
	// returns true if filename has .bwt extension.
	int n = (int)strlen(fileName);
	if (n < 4)
		return false;
	if (strncmp(&fileName[n - 4], ".bwt", 4) == 0)
		return true;
	return false;
}





void CreateBwtFileName(char * inputFileName, char * & bwtFileName)
{
	// create output bwt file name from original input file name.
	int inLen = (int)strlen(inputFileName);
	int n = inLen;
	while (--n > 0)
	{
		switch (inputFileName[n])
		{
			case '.':
			{
				bwtFileName = new char[n + 5];
				strncpy(bwtFileName, inputFileName, n + 1);
				bwtFileName[n + 1] = 'b';
				bwtFileName[n + 2] = 'w';
				bwtFileName[n + 3] = 't';
				bwtFileName[n + 4] = 0;
				return;
			}
	
			case '\\':
			case '/':
				break;
		}
	}

	bwtFileName = new char[inLen + 5];
	sprintf(bwtFileName, "%s.bwt", inputFileName);
}





int CreateOriginalFileName(char * bwtFileName, char * bwtFile, char * & outputFileName)
{
	// creates the original file name from which the bwt file came.
	int inLen = (int)strlen(bwtFileName);
	int originalExtLen = 0;
	while (bwtFile[originalExtLen] != 0)
		originalExtLen++;

	outputFileName = new char[2048];
	int p = (int)(strlen(bwtFileName) - 4);
	strncpy(outputFileName, bwtFileName, p);
	strncpy(&outputFileName[p], bwtFile, originalExtLen);

	outputFileName[p + originalExtLen] = 0;

	return originalExtLen + 1;
}








void ForwardBWT(char * file)
{
	// test the MSufSort algorithm for use with the burrows wheeler transform
	char * sourceBuffer;
	int size;
	BWT bwt;

	printf("\n\nProcessing: %s\n", GetFileName(file));

	LoadSourceFile(file, sourceBuffer, size);
	if (sourceBuffer == 0)
	{
		printf("Invalid source or source of zero length\n");
		return;
	}

	// Call forward BWT
	unsigned int key = bwt.Forward((unsigned char *)sourceBuffer, size);

	// save the BWT to file.
	char * outputFile;
	CreateBwtFileName(file, outputFile);
	FILE * outFP = fopen(outputFile, "wb");

	// output original file ext. (zero terminated)
	int p = (int)strlen(file);
	int n = ((int)strlen(outputFile) - 4);
	if ((p - n) > 0)
		fwrite(&file[n], 1, p - n, outFP);
	fputc(0, outFP);

	// output the key needed to reverse the BWT.
	fwrite((char *)&key, 1, 4, outFP);
	
	// output the BWT
	fwrite(sourceBuffer, 1, size, outFP);
	fclose(outFP);

	// clean up and done.
	delete [] outputFile;
	delete [] sourceBuffer;
}




void ReverseBWT(char * file)
{
	// undo the bwt file created with TestForwardBWT()
	char * sourceBuffer;
	int size;

	LoadSourceFile(file, sourceBuffer, size);
	if (sourceBuffer == 0)
	{
		printf("Invalid source or source of zero length\n");
		return;
	}

	BWT bwt;
	// create the output file name.
	char * outputFileName = 0;
	int keyOffset = CreateOriginalFileName(file, sourceBuffer, outputFileName);

	// get the reverse bwt 32 bit key.
	int * key = (int *)&sourceBuffer[keyOffset];

	printf("\n\nProcessing: %s\n", file);
	
	// Call reverse BWT
	size -= (keyOffset + 4);
	bwt.Reverse((unsigned char *)&sourceBuffer[(keyOffset + 4)], size, *key);

	// write out reversed bwt.
	FILE * outFP = fopen(outputFileName, "wb");
	fwrite(&sourceBuffer[(keyOffset + 4)], 1, size, outFP);
	fclose(outFP);

	delete [] outputFileName;
	delete [] sourceBuffer;
}






int main(int argc, char * argv[])
{
	for (int i = 1; i < argc; i++)
	{
		if (IsBwtFile(argv[i]))
			ReverseBWT(argv[i]);
		else
			ForwardBWT(argv[i]);
	}

	printf("All tasks completed.\n");
	getch();
}

