#ifndef MSUFSORT_STACK_H
#define MSUFSORT_STACK_H

//=============================================================================================
// A quick and dirty stack class for use with the MSufSort algorithm
//
// Author: M.A. Maniscalco
// Date: 7/30/04
// email: michael@www.michael-maniscalco.com
//
// This code is free for non commercial use only.
//
//=============================================================================================

#include "memory.h"


template <class T>
class Stack 
{
public:
	Stack(unsigned int startingSize, unsigned int expandingSize, bool preAllocate = false):
			m_startingSize(startingSize), m_expandingSize(expandingSize), m_preAllocate(preAllocate)
	{
		Initialize();
	}

	virtual ~Stack(){SetSize(0);}

	void Push(T & value);

	T & Pop();

	void SetSize(unsigned int stackSize);

	void Initialize();

	unsigned int Count();

	void Clear();

	T *				m_stack;

	T *				m_stackPtr;

	T *				m_endOfStack;

	unsigned int	m_stackSize;

	unsigned int	m_startingSize;

	unsigned int	m_expandingSize;

	bool			m_preAllocate;
};






template <class T> 
inline void Stack<T>::Clear()
{
	m_stackPtr = m_stack;	
}




template <class T> 
inline unsigned int Stack<T>::Count()
{
	return (unsigned int)(m_stackPtr - m_stack);
}




template <class T> 
inline void Stack<T>::Initialize()
{
	m_stack = m_endOfStack = m_stackPtr = 0;
	m_stackSize = 0;
	if (m_preAllocate)
		SetSize(m_startingSize);
}




template <class T> 
inline void Stack<T>::Push(T & value)
{
	if (m_stackPtr >= m_endOfStack)
		SetSize((m_stackSize == 0) ? m_startingSize : m_stackSize + m_expandingSize);

	*(m_stackPtr++) = value;
}






template <class T> 
inline T & Stack<T>::Pop()
{
	return *(--m_stackPtr);
}





template <class T> 
inline void Stack<T>::SetSize(unsigned int stackSize)
{
	if (m_stackSize == stackSize)
		return;

	T * newStack = 0;
	if (stackSize)
	{
		newStack = new T[stackSize];
		unsigned int bytesToCopy = (unsigned int)(m_stackPtr - m_stack) * (unsigned int)sizeof(T);
		if (bytesToCopy)
			memcpy(newStack, m_stack, bytesToCopy);

		m_stackPtr = &newStack[m_stackPtr - m_stack];
		m_endOfStack = &newStack[stackSize];
		m_stackSize = stackSize;
	}

	if (m_stack)
		delete [] m_stack;
	m_stack = newStack;
}





#endif